
## constants.py
## used to define constants that are used in multiple places in the package
## these constants should not be exported



class Constants:

    MALE_SYMBOL = 'MALE'
    FEMALE_SYMBOL = 'FEMALE'
    OTHER_SEX_SYMBOL = 'OTHER'
    UNKOWN_SEX_SYMBOL = 'UNKNOWN'
    NOT_PROVIDED = 'NOT_PROVIDED'

